using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.IO;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Windows.Forms;

using Smdn.Formats.Ini;
using Smdn.Formats.Mime;
using Smdn.Protocols.Imap4;
using Smdn.Protocols.Imap4.Client;

namespace Smdn.Applications.TundereBird {
  public class MainWindow : Form {
    public MainWindow()
    {
      agent = new ImapClientAgent();
      agent.Initialize();
      agent.ServerCertificateValidationCallback += ValidateServerCertificate;

      InitializeComponents();
    }

    protected override void OnLoad (EventArgs e)
    {
      base.OnLoad (e);
    }

    protected override void OnShown (EventArgs e)
    {
      base.OnShown (e);

      Application.DoEvents();

      LoadConfig();

      InitializeMailboxTree();
    }

    protected override void OnClosing (CancelEventArgs e)
    {
      base.OnClosing (e);

      if (!e.Cancel) {
        LogoutAll();

        agent.Dispose();
      }
    }

    private bool ValidateServerCertificate(object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors)
    {
      if (sslPolicyErrors == SslPolicyErrors.None)
        return true;

      var message = new StringBuilder();

      message.AppendLine("サーバ証明書に問題があります。　続行しますか?");
      message.AppendLine();
      message.AppendFormat("SslPolicyErrors: {0}", sslPolicyErrors);
      message.AppendLine();
      message.AppendFormat("X509Certificate: {0}{1}", Environment.NewLine, certificate.ToString(false));
      message.AppendLine();

      if (chain != null) {
        foreach (var element in chain.ChainElements) {
          message.AppendFormat("発行先: {0}", element.Certificate.Subject);
          message.AppendLine();
          message.AppendFormat("発行者: {0}", element.Certificate.Issuer);
          message.AppendLine();
          message.AppendFormat("有効期間: {0} から {1}", element.Certificate.GetEffectiveDateString(), element.Certificate.GetExpirationDateString());
          message.AppendLine();

          foreach (var status in element.ChainElementStatus) {
            message.AppendFormat("Status: {0}", status.Status);
            message.AppendLine();
            message.AppendFormat("StatusInformation: {0}", status.StatusInformation);
            message.AppendLine();
            message.AppendLine();
          }
        }
      }

      var result = MessageBox.Show(message.ToString(),
                                   "ValidateServerCertificate",
                                   MessageBoxButtons.YesNo);

      return (result == DialogResult.Yes);
    }

    private void LoadConfig()
    {
      var config = Document.Load("tunderebird.ini");
      var clients = new List<ImapClient>();

      foreach (var section in config) {
        if (string.Empty.Equals(section.Name)) {
          continue;
        }
        else if (section.Name == "CertFiles") {
          // 証明書ファイルのセクション
          foreach(var entry in section) {
            try {
              agent.CertCollection.Add(new X509Certificate2(entry.Value));
            }
            catch (Exception ex) {
              AppendStatus("証明書{0}を読み込めませんでした", entry.Value);
              AppendStatus("Exception: {0}", ex);
            }
          }
        }
        else {
          // アカウントのセクション
          try {
            var client = agent.CreateClient();

            client.HostUri = new ImapUri(section["HostURL"]);
            client.Password = section["Password"];
            client.UseSsl = (section["UseSSL"] == "yes");

            clients.Add(client);
          }
          catch (Exception ex) {
            AppendStatus("{0}のアカウント設定を読み込めませんでした", section.Name);
            AppendStatus("Exception: {0}", ex);
          }
        }
      }

      this.clients = clients.ToArray();
    }

    private void InitializeMailboxTree()
    {
      foreach (var client in clients) {
        // ログインする
        try {
          AppendStatus("{0}へ接続しています", client.HostUri);

          client.Login();
        }
        catch (Exception ex) {
          AppendStatus("{0}へ接続できませんでした", client.HostUri);
          AppendStatus("Exception: {0}", ex);
          continue;
        }

        // メールボックスの階層構造を取得する
        ImapMailboxHierarchy[] hierarchies;

        try {
          AppendStatus("{0}のメールボックスを取得しています", client.HostUri);

          hierarchies = client.GetMailboxHierarchies();
        }
        catch (Exception ex) {
          AppendStatus("{0}のメールボックス一覧を取得できませんでした", client.HostUri);
          AppendStatus("Exception: {0}", ex);
          continue;
        }

        var mailboxList = new List<ImapMailbox>();

        foreach (ImapMailboxHierarchy hierarchy in hierarchies) {
          foreach (ImapMailboxHierarchy h in hierarchy.Flatten()) {
            mailboxClientMap.Add(h.Mailbox, client);
            mailboxList.Add(h.Mailbox);
          }
        }

        // アカウントとメールボックスのノードを作成する
        try {
          mailboxTree.BeginUpdate();

          var rootNode = new ClientNode(client);

          CreateMailboxNode(hierarchies, rootNode);

          rootNode.ExpandAll();

          mailboxTree.Nodes.Add(rootNode);
        }
        finally {
          mailboxTree.EndUpdate();
        }

        mailboxTree.Refresh();

        // メールボックスの状態を更新
        AppendStatus("{0}のメールボックスの状態を取得しています", client.HostUri);

        foreach (var mailbox in mailboxList) {
          try {
            client.UpdateMailboxStatus(mailbox);

            mailboxNodeMap[mailbox].UpdateText();
            mailboxTree.Refresh();
          }
          catch (Exception ex) {
            AppendStatus("{0}の情報を取得できませんでした", mailbox.Uri);
            AppendStatus("Exception: {0}", ex);
          }
        }

        AppendStatus("完了");
      }
    }

    private void CreateMailboxNode(ImapMailboxHierarchy[] hierarchies, TreeNode parentNode)
    {
      foreach (ImapMailboxHierarchy hierarchy in hierarchies) {
        var node = new MailboxNode(hierarchy.Name, hierarchy.Mailbox);

        parentNode.Nodes.Add(node);

        CreateMailboxNode(hierarchy.SubHierarchies, node);

        mailboxNodeMap.Add(hierarchy.Mailbox, node);
      }
    }

    private void ClientOrMailboxSelected(object sender, TreeNodeMouseClickEventArgs e)
    {
      if (!(e.Node is MailboxNode))
        return;

      var mailbox = (e.Node as MailboxNode).Mailbox;

      // メッセージの一覧を取得
      AppendStatus("{0}のメッセージの一覧を取得しています", mailbox.Uri);

      try {
        var messages = mailboxClientMap[mailbox].ListMessages(mailbox);

        CreateMessageList(messages);

        AppendStatus("全{0}件のメッセージがあります", messages.Length);
      }
      catch (Exception ex) {
        AppendStatus("{0}のメッセージ一覧を取得できませんでした", mailbox.Uri);
        AppendStatus("Exception: {0}", ex);
      }
    }

    private void CreateMessageList(ImapMessage[] messages)
    {
      try {
        messageListView.BeginUpdate();
        messageListView.Items.Clear();

        foreach (var message in messages) {
          messageListView.Items.Add(new MessageItem(message));
        }
      }
      finally {
        messageListView.EndUpdate();
      }
    }

    private void MessageSelected(object sender, EventArgs e)
    {
      if (messageListView.SelectedItems.Count <= 0)
        return;

      var messageItem = (messageListView.SelectedItems[0] as MessageItem); 

      if (messageItem == null)
        return;

      AppendStatus("{0}の本文を取得しています", messageItem.Message.Uri);

      var message = messageItem.Message;
      ImapMessage messageBody = null;

      try {
        messageBody = mailboxClientMap[message.Mailbox].FetchMessage(message);
      }
      catch (Exception ex) {
        AppendStatus("{0}の本文を取得できませんでした", message.Uri);
        AppendStatus("Exception: {0}", ex);
        return;
      }

      try {
        var body = MimeMessage.Create(new StringReader(messageBody.Body));

        if (body.Body == null)
          messageView.Text = "(非テキストメッセージです)";
        else
          messageView.Text = body.Body;

        if (0 < body.SubParts.Count) {
          messageView.Text += "-- multipart message --";
          messageView.Text += Environment.NewLine;

          foreach (var subPart in body.SubParts) {
            if (subPart.Body == null)
              messageView.Text = "(非テキストメッセージです)";
            else
              messageView.Text = subPart.Body;

            messageView.Text += Environment.NewLine;
            messageView.Text += "-- end of part --";
          }
        }
      }
      catch (Exception ex) {
        AppendStatus("{0}の解析に失敗しました", message.Uri);
        AppendStatus("Exception: {0}", ex);
      }

      AppendStatus("完了");
    }

    private void NoOp()
    {
      foreach (var client in clients) {
        client.NoOp();
      }
    }

    private void LogoutAll()
    {
      foreach (var client in clients) {
        try {
          client.Logout();
        }
        catch (Exception ex) {
          AppendStatus("{0}からログアウト出来ませんでした", client.HostUri);
          AppendStatus("Exception: {0}", ex);
        }
      }
    }

    private void AppendStatus(string format, params object[] args)
    {
      statusView.Text += string.Format(format, args);
      statusView.Text += Environment.NewLine;
      statusView.Select(statusView.TextLength - 1, 0);
      statusView.ScrollToCaret();
      statusView.Update();
    }

    private void InitializeComponents()
    {
      this.Text = "TundereBird";
      this.ClientSize = new Size(640, 480);
      this.StartPosition = FormStartPosition.CenterScreen;

      mailboxTree = new TreeView();
      mailboxTree.Location = new Point(0, 0);
      mailboxTree.Size = new Size(200, 480);
      mailboxTree.Anchor = AnchorStyles.Left | AnchorStyles.Top | AnchorStyles.Bottom;
      mailboxTree.ShowLines = true;
      mailboxTree.ShowPlusMinus = true;
      mailboxTree.NodeMouseClick += ClientOrMailboxSelected;

      this.Controls.Add(mailboxTree);

      messageListView = new ListView();
      messageListView.Location = new Point(200, 0);
      messageListView.Size = new Size(440, 160);
      messageListView.Anchor = AnchorStyles.Left | AnchorStyles.Right | AnchorStyles.Top;
      messageListView.View = View.Details;
      messageListView.Font = SystemFonts.IconTitleFont;
      messageListView.Columns.Add("タイトル", 200);
      messageListView.Columns.Add("サイズ", 40);
      messageListView.Columns.Add("受信日", 60);
      messageListView.Columns.Add("送信者", 100);
      messageListView.Columns.Add("フラグ", 200);
      messageListView.SelectedIndexChanged += MessageSelected;

      this.Controls.Add(messageListView);

      messageView = new TextBox();
      messageView.Location = new Point(200, 160);
      messageView.Size = new Size(440, 260);
      messageView.Anchor = AnchorStyles.Left | AnchorStyles.Right | AnchorStyles.Top | AnchorStyles.Bottom;
      messageView.Multiline = true;
      messageView.Font = SystemFonts.DefaultFont;
      messageView.ScrollBars = ScrollBars.Both;

      this.Controls.Add(messageView);

      statusView = new TextBox();
      statusView.Location = new Point(200, 420);
      statusView.Size = new Size(440, 60);
      statusView.Anchor = AnchorStyles.Left | AnchorStyles.Right | AnchorStyles.Bottom;
      statusView.ReadOnly = true;
      statusView.Multiline = true;
      statusView.Font = SystemFonts.DefaultFont;
      statusView.ScrollBars = ScrollBars.Both;

      this.Controls.Add(statusView);

      timerUpdate = new Timer();
      timerUpdate.Enabled = true;
      timerUpdate.Interval = 5 * 60 * 1000; // 5 min
      timerUpdate.Tick += delegate {
        NoOp();
      };
    }


    private class ClientNode : TreeNode {
      public ImapClient Client {
        get { return client; }
      }

      public ClientNode(ImapClient client)
      {
        this.client = client;

        Text = client.HostUri.UserAtHost;
      }

      private readonly ImapClient client;
    }

    private class MailboxNode : TreeNode {
      public ImapMailbox Mailbox {
        get { return mailbox; }
      }

      public MailboxNode(string name, ImapMailbox mailbox)
      {
        this.name = name;
        this.mailbox = mailbox;

        UpdateText();
      }

      public void UpdateText()
      {
        Text = string.Format("{0} 新着{1} 未読{2} 全{3}",
                             name,
                             mailbox.RecentMessage,
                             mailbox.UnseenMessage,
                             mailbox.ExistsMessage);
      }

      private readonly string name;
      private readonly ImapMailbox mailbox;
    }

    private class MessageItem : ListViewItem {
      public ImapMessage Message {
        get { return message; }
      }

      public MessageItem(ImapMessage message)
      {
        this.message = message;

        if (message.Envelope == null) {
          this.Text = string.Empty;
          this.SubItems.Add(message.BodySize.ToString());
          this.SubItems.Add(string.Empty);
          this.SubItems.Add(string.Empty);
          this.SubItems.Add(message.Flags.ToString());
        }
        else {
          this.Text = Smdn.Formats.Mime.Decoder.DecodeMimeEncodedString(message.Envelope.Subject ?? string.Empty);
          this.SubItems.Add(message.BodySize.ToString());
          this.SubItems.Add(message.Envelope.Date.ToString());
          this.SubItems.Add(FromAddress(message.Envelope.From));
          this.SubItems.Add(message.Flags.ToString());
        }
      }

      private string FromAddress(ImapAddress address)
      {
        if (address == null)
          return string.Empty;

        return string.Format("{0} <{1}@{2}>",
                             Smdn.Formats.Mime.Decoder.DecodeMimeEncodedString(address.Name ?? string.Empty),
                             address.Mailbox,
                             address.Host);
      }

      private readonly ImapMessage message;
    }

    private ImapClientAgent agent;
    private ImapClient[] clients;
    private readonly Dictionary<ImapMailbox, MailboxNode> mailboxNodeMap = new Dictionary<ImapMailbox, MailboxNode>();
    private readonly Dictionary<ImapMailbox, ImapClient> mailboxClientMap = new Dictionary<ImapMailbox, ImapClient>();

    private TreeView mailboxTree;
    private ListView messageListView;
    private TextBox messageView;
    private TextBox statusView;
    private Timer timerUpdate;
  }
}
