﻿using System;
using System.Collections.Generic;
using System.Web;
using System.Text;

using Smdn.Protocols.Imap4.Transaction;

namespace Smdn.Protocols.Imap4 {
  // RFC 2192 IMAP URL Scheme
  //   http://tools.ietf.org/html/rfc2192
  //   http://www.lins.jp/~obata/imap/rfc/rfc2192ja.html
  
  // An IMAP URL takes one of the following forms:
  //   imap://<iserver>/
  //   imap://<iserver>/<enc_list_mailbox>;TYPE=<list_type>
  //   imap://<iserver>/<enc_mailbox>[uidvalidity][?<enc_search>]
  //   imap://<iserver>/<enc_mailbox>[uidvalidity]<iuid>[isection]

  public class ImapUri : Uri {
    public const string UriSchemeImap = "imap";
    internal const int DefaultPort = 143;

    /// <summary>
    /// returns username without authentication type
    /// </summary>
    public string UserName {
      get
      {
        if (UserInfo.Contains(";"))
          return UserInfo.Substring(0, UserInfo.IndexOf(";"));
        else
          return UserInfo;
      }
    }

    /// <summary>
    /// returns username@host
    /// </summary>
    public string UserAtHost {
      get { return string.Format("{0}@{1}", UserName, Host); }
    }

    public ImapAuthenticationType Authentication {
      get
      {
        var index = UserInfo.ToUpper().IndexOf(";AUTH=");

        if (index == -1)
          return ImapAuthenticationType.None;
        else
          return ImapConverter.ToAuthType(UserInfo.Substring(index + 6));
      }
    }

    public string ListMailbox {
      get
      {
        var index = AbsolutePath.ToUpper().IndexOf(";TYPE=");

        if (0 <= index)
          return AbsolutePath.Substring(1, index - 1); 
        else
          return string.Empty;
      }
    }

    public ImapListType ListType {
      get
      {
        var index = AbsolutePath.ToUpper().IndexOf(";TYPE=");

        if (index == -1)
          return ImapListType.None;
        else
          return ImapConverter.ToListType(AbsolutePath.Substring(index + 6));
      }
    }

    public string Mailbox {
      get
      {
        var mailbox = AbsolutePath.Substring(1);

        for (;;) {
          var start = mailbox.IndexOf(';');

          if (start == -1)
            break;

          var end = mailbox.IndexOf('/', start);

          if (end == -1)
            mailbox = mailbox.Substring(0, start);
          else
            mailbox = mailbox.Substring(0, start) + mailbox.Substring(start, end - start);
        }

        if (mailbox.EndsWith("/"))
          mailbox = mailbox.Substring(0, mailbox.Length - 1);

        return HttpUtility.UrlDecode(mailbox);
      }
    }

    public int UidValidity {
      get
      {
        var path = AbsolutePath;

        var start = path.ToUpper().IndexOf(";UIDVALIDITY=");

        if (start == -1)
          return 0;

        var validity = path.Substring(start + 13);

        var end = validity.IndexOf('/');

        if (end == -1)
          return int.Parse(validity);
        else
          return int.Parse(validity.Substring(0, end));
      }
    }

    public int MessageUid {
      get
      {
        var path = AbsolutePath;

        var start = path.ToUpper().IndexOf("/;UID=");

        if (start == -1)
          return 0;

        var uid = path.Substring(start + 6);

        var end = uid.IndexOf('/');

        if (end == -1)
          return int.Parse(uid);
        else
          return int.Parse(uid.Substring(0, end));
      }
    }

    public string Section {
      get
      {
        var path = AbsolutePath;

        var start = path.ToUpper().IndexOf("/;SECTION=");

        if (start == -1)
          return string.Empty;

        var section = path.Substring(start + 10);

        var end = section.IndexOf('/');

        if (end == -1)
          return section;
        else
          return section.Substring(0, end);
      }
    }

    internal ImapUriForm Form {
      get
      {
        if (MessageUid != 0)
          return ImapUriForm.FetchMessage;
        else if (!string.Empty.Equals(Query))
          return ImapUriForm.SearchMessages;
        else if (ListType != ImapListType.None)
          return ImapUriForm.ListMailboxes;
        else if (!string.Empty.Equals(Mailbox))
          return ImapUriForm.ListMessages;
        else
          return ImapUriForm.Server;
      }
    }

    public ImapUri(Uri uri)
      : this(uri.ToString())
    {
    }

    public ImapUri(string uriString)
      : base(uriString)
    {
      if (Scheme != UriSchemeImap)
        throw new ArgumentException(string.Format("scheme must be {0}", UriSchemeImap));
    }

    public ImapUri(ImapUri baseUri, string relativeUri)
      : base(baseUri, relativeUri)
    {
    }

    public ImapUri(ImapUri baseUri, Uri relativeUri)
      : base(baseUri, relativeUri)
    {
    }

    public ImapUri GetParent()
    {
      var authority = this.GetLeftPart(UriPartial.Authority);
      var path = this.AbsolutePath;
      int sepPos;

      if (path == "/")
        return new ImapUri(authority);

      if (path.EndsWith("/"))
        sepPos = path.LastIndexOf('/', path.Length - 2);
      else
        sepPos = path.LastIndexOf('/');

      if (sepPos == -1)
        return new ImapUri(authority);
      else
        path = path.Substring(0, sepPos + 1);

      return new ImapUri(new ImapUri(authority), path);
    }
  }
}
