using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Net.Security;
using System.IO;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;
using System.Threading;
using System.Web;

using Smdn.Protocols.Imap4.Transaction;

namespace Smdn.Protocols.Imap4.Session {
  // Layers:
  //   Client.ImapClient
  //   Client.ImapClientAgent
  // * Session.ImapSessionAgent
  //   Session.ImapSession
  //   Transaction.ImapTransactionAgent
  //   Transaction.ImapTransaction/Transaction.ImapRequest/Transaction.ImapResponse
  //   Protocol.ImapConnection/Protocol.ImapProtocolCommand/Protocol.ImapProtocolResponse
  //   (IMAP4)

  public class ImapSessionAgent : IDisposable {
    public X509Certificate2Collection CertCollection {
      get { return certCollection; }
    }

    public RemoteCertificateValidationCallback ServerCertificateValidationCallback {
      get { return serverCertificateValidationCallback; }
      set { serverCertificateValidationCallback = value; }
    }

    public ImapCreateAuthenticatedStreamCallback CreateAuthenticatedStreamCallback {
      get { return createAuthenticatedStreamCallback; }
      set { createAuthenticatedStreamCallback = value; }
    }

    /*
     * construction/destruction
     */

    public ImapSessionAgent()
    {
    }

    public void Initialize()
    {
      if (agent != null)
        throw new InvalidOperationException("already initialized");

      Trace.Info("ImapSessionAgent initializing");

      agent = new ImapTransactionAgent();
      agent.Initialize();

      Trace.Info("ImapSessionAgent initialized");
    }

    ~ImapSessionAgent()
    {
      Dispose();
      GC.SuppressFinalize(this);
    }

    public void Dispose()
    {
      if (disposed)
        return;

      Trace.Info("ImapSessionAgent disposing");

      if (agent != null) {
        List<ImapSession> existSessions;

        lock (sessions) {
          existSessions = new List<ImapSession>(sessions.Values);
        }

        foreach (var session in existSessions) {
          if (session.Status == ImapSessionStatus.Selected)
            Close(session);
          if (session.Status == ImapSessionStatus.Authenticated)
            Logout(session);
          if (session.Status != ImapSessionStatus.NotConnected)
            Disconnect(session);
        }

        lock (sessions) {
          sessions.Clear();
        }

        agent.Dispose();
        agent = null;
      }

      Trace.Info("ImapSessionAgent disposed");

      disposed = true;
    }

    /*
     * methods for requesting by IMAP URL
     */

    /// <summary>sends request by IMAP URL</summary>
    public bool Request(ImapSession session, ImapUri requestUri, out ImapResponse response)
    {
      response = null;

      RejectNonConnectedState(session);

      if (session != GetExistSession(requestUri))
        throw new InvalidOperationException("the host specified with IMAP URL is different from the session");

      TraceVerbose(session, "requesting by IMAP URL '{0}'", requestUri);

      switch (requestUri.Form) {
        case ImapUriForm.Server:
        case ImapUriForm.ListMailboxes:
          return ListMailboxes(session, requestUri, out response);

        case ImapUriForm.ListMessages:
          return ListMessages(session, requestUri, out response);

        case ImapUriForm.FetchMessage:
          return FetchMessage(session, requestUri, out response);

        case ImapUriForm.SearchMessages:
          return SearchMessages(session, requestUri, out response);

        default:
          throw new UriFormatException();
      }
    }

    private bool ListMailboxes(ImapSession session, ImapUri requestUri, out ImapResponse response)
    {
      RejectNonAuthenticatedState(session);

      response = null;

      ImapMailbox[] mailboxes;
      ImapTransactionMethod listMethod;

      switch (requestUri.ListType) {
        case ImapListType.List:
          listMethod = ImapTransactionMethod.List;
          break;
        case ImapListType.Lsub:
          listMethod = ImapTransactionMethod.Lsub;
          break;
        default:
          listMethod = ImapTransactionMethod.Lsub;
          break;
      }

      if (!ListLsubCore(listMethod, session, requestUri.Mailbox, out mailboxes))
        return false;

      response = new ImapListResponse(ImapTransactionStatus.Ok, mailboxes, "List succeeded");

      return true;
    }

    private bool ListMessages(ImapSession session, ImapUri requestUri, out ImapResponse response)
    {
      /*
      RFC 2683 IMAP4 Implementation Recommendations
      http://tools.ietf.org/html/rfc2683

      3.2.1.2. Fetching the List of Messages
         When a client selects a mailbox, it is given a count, in the untagged
         EXISTS response, of the messages in the mailbox.  This number can be
         very large.  In such a case it might be unwise to use

             C: 004 FETCH 1:* ALL

         to populate the user's view of the mailbox.  One good method to avoid
         problems with this is to batch the requests, thus:

             C: 004 FETCH 1:50 ALL
             S: * 1 FETCH ...etc...
             S: 004 OK done
             C: 005 FETCH 51:100 ALL
             S: * 51 FETCH ...etc...
             S: 005 OK done
             C: 006 FETCH 101:150 ALL
             ...etc...
      */

      response = null;

      var selectedMailbox = session.SelectedMailbox;

      if (!Status(session, selectedMailbox))
        return false;

      var messageCount = selectedMailbox.ExistsMessage;
      var fetchedMessages = new List<ImapMessage>();

      for (var fetching = 1; fetching <= messageCount; ) {
        const int fetchUpTo = 20;
        string fetchingSet;

        if (fetching + fetchUpTo <= messageCount)
          fetchingSet = string.Format("{0}:{1}", fetching, fetching + fetchUpTo - 1);
        else
          fetchingSet = string.Format("{0}:*", fetching);

        fetching += fetchUpTo;

        ImapMessage[] messages;

        if (!Fetch(session, fetchingSet, "(uid flags internaldate rfc822.size envelope body)", out messages))
          return false;

        fetchedMessages.AddRange(messages);
      }

      response = new ImapMessageResponse(ImapTransactionStatus.Ok, fetchedMessages.ToArray(), "Fetch succeeded"); 

      return true;
    }

    private bool FetchMessage(ImapSession session, ImapUri requestUri, out ImapResponse response)
    {
      /*
      RFC 2683 IMAP4 Implementation Recommendations
      http://tools.ietf.org/html/rfc2683

      3.2.1.3. Fetching a Large Body Part

         The issue here is similar to the one for a list of messages.  In the
         BODYSTRUCTURE response the client knows the size, in bytes, of the
         body part it plans to fetch.  Suppose this is a 70 MB video clip. The
         client can use partial fetches to retrieve the body part in pieces,
         avoiding the problem of an uninterruptible 70 MB literal coming back
         from the server:

             C: 022 FETCH 3 BODY[1]<0.20000>
             S: * 3 FETCH (FLAGS(\Seen) BODY[1]<0> {20000}
             S: ...data...)
             S: 022 OK done
             C: 023 FETCH 3 BODY[1]<20001.20000>
             S: * 3 FETCH (BODY[1]<20001> {20000}
             S: ...data...)
             S: 023 OK done
             C: 024 FETCH 3 BODY[1]<40001.20000>
             ...etc...
      */

      response = null;

      ImapMessage[] messages;
      var dataItems = string.Format("(uid flags internaldate rfc822.size envelope body body.peek[{0}])", requestUri.Section);

      if (!UidFetch(session, new int[] {requestUri.MessageUid}, dataItems, out messages))
        return false;

      response = new ImapMessageResponse(ImapTransactionStatus.Ok, messages, "Fetch succeeded"); 

      return true;
    }

    private bool SearchMessages(ImapSession session, ImapUri requestUri, out ImapResponse response)
    {
      response = null;

      // TODO: 非ASCII文字の取扱い
      var query = HttpUtility.UrlDecode(requestUri.Query.Substring(1));
      int[] messageSequences;

      if (!Search(session, query, out messageSequences))
        return false;

      // RFC 2683 IMAP4 Implementation Recommendations
      // http://tools.ietf.org/html/rfc2683
      // 3.2.1.2. Fetching the List of Messages
      var fetchedMessages = new List<ImapMessage>();

      for (var fetching = 0; fetching <= messageSequences.Length; ) {
        const int fetchUpTo = 20;
        var fetchingSet = string.Empty;

        for (var i = 0; i < fetchUpTo; i++) {
          if (!string.Empty.Equals(fetchingSet))
            fetchingSet += ",";
          fetchingSet = messageSequences[fetching++].ToString();
          if (fetching <= messageSequences.Length)
            break;
        }

        ImapMessage[] messages;

        if (!Fetch(session, fetchingSet, "(uid flags internaldate rfc822.size envelope body)", out messages))
          return false;

        fetchedMessages.AddRange(messages);
      }

      response = new ImapMessageResponse(ImapTransactionStatus.Ok, fetchedMessages.ToArray(), "Search succeeded");

      return true;
    }

    /*
     * method for connect
     */

    /// <summary>connects to IMAP server</summary>
    public bool Connect(ImapUri uri, out ImapSession session)
    {
      session = GetExistSession(uri);

      if (session != null)
        return true;

      session = ImapSession.Start(uri, CreateAuthenticatedStream);

      if (session == null)
        return false;

      TraceVerbose(session, "waiting for initial server greeting response");

      ImapResponse response;

      if (RequestSync(session, ImapTransactionMethod.Greeting, out response)) {
        lock (sessions) {
          sessions.Add(uri.UserAtHost, session);
        }

        if (response.Status == ImapTransactionStatus.PreAuthed)
          session.Status = ImapSessionStatus.Authenticated;
        else
          session.Status = ImapSessionStatus.NonAuthenticated;

        return true;
      }
      else {
        session = null;

        return false;
      }
    }

    /*
     * requesting methods for any state
     */

    /// <summary>sends CAPABILITY command</summary>
    /// <remarks>valid in any state</remarks>
    public bool Capability(ImapSession session, out ImapCapability capability)
    {
      RejectNonConnectedState(session);

      capability = null;

      ImapResponse response;

      if (RequestSync(session, ImapTransactionMethod.Capability, out response)) {
        var resp = (response as ImapCapabilityResponse);

        capability = resp.Capability;

        return true;
      }

      return false;
    }

    /// <summary>sends NOOP command</summary>
    /// <remarks>valid in any state</remarks>
    public bool NoOp(ImapSession session)
    {
      RejectNonConnectedState(session);

      return RequestSync(session, ImapTransactionMethod.NoOp);
    }

    /// <summary>sends LOGOUT command</summary>
    /// <remarks>valid in any state</remarks>
    public bool Logout(ImapSession session)
    {
      if (session.Status == ImapSessionStatus.NotConnected) {
        TraceVerbose(session, "already logged out or disconnected");
        return true;
      }

      if (session.Status == ImapSessionStatus.NonAuthenticated) {
        TraceVerbose(session, "not logged in");
        return true;
      }

      ImapResponse response;

      RequestSync(session, ImapTransactionMethod.Logout, out response);

      if (response.Status == ImapTransactionStatus.Ok) {
        session.Status = ImapSessionStatus.NonAuthenticated;

        TraceVerbose(session, "now in non-authenticated state");

        Disconnect(session);

        return true;
      }
      else if (response.Status == ImapTransactionStatus.Disconnecting) {
        return true;
      }
      else {
        return false;
      }
    }

    /*
     * requesting methods for non authenticated state
     */

    /// <summary>sends STARTTLS command</summary>
    /// <remarks>valid in non authenticated state</remarks>
    public bool StartTls(ImapSession session)
    {
      RejectNonConnectedState(session);

      throw new NotImplementedException("use Login() or Authenticate()");
    }

    /// <summary>sends AUTHENTICATE command</summary>
    /// <remarks>valid in non authenticated state</remarks>
    public bool Authenticate(ImapSession session, string password)
    {
      RejectNonConnectedState(session);

      if (session.Status != ImapSessionStatus.NonAuthenticated) {
        TraceVerbose(session, "already authenticated");
        return true;
      }

      var request = new ImapRequest(ImapTransactionMethod.Authenticate, session.Timeout);

      request.Parameters["Username"] = session.HostUri.UserName;
      request.Parameters["Password"] = password;

      switch (session.HostUri.Authentication) {
        case ImapAuthenticationType.None:
        case ImapAuthenticationType.Any:
          request.Parameters["Auth-Type"] = string.Empty;
          break;
        default:
          request.Parameters["Auth-Type"] = ImapConverter.ToString(session.HostUri.Authentication);
          break;
      }

      if (RequestSync(session, request)) {
        session.Status = ImapSessionStatus.Authenticated;

        TraceVerbose(session, "now in authenticated state");

        return true;
      }

      return false;
    }

    /// <summary>sends LOGIN command</summary>
    /// <remarks>valid in non authenticated state</remarks>
    public bool Login(ImapSession session, string password)
    {
      RejectNonConnectedState(session);

      if (session.Status != ImapSessionStatus.NonAuthenticated) {
        TraceVerbose(session, "already logged in");
        return true;
      }

      var request = new ImapRequest(ImapTransactionMethod.Login, session.Timeout);

      request.Parameters["Username"] = session.HostUri.UserName;
      request.Parameters["Password"] = password;

      if (RequestSync(session, request)) {
        session.Status = ImapSessionStatus.Authenticated;

        TraceVerbose(session, "now in authenticated state");

        return true;
      }

      return false;
    }

    /*
     * requesting methods for authenticated state
     */

    /// <summary>sends SELECT command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Select(ImapSession session, ImapMailbox mailbox)
    {
      return SelectExamineCore(ImapTransactionMethod.Select, session, mailbox);
    }

    /// <summary>sends EXAMINE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Examine(ImapSession session, ImapMailbox mailbox)
    {
      return SelectExamineCore(ImapTransactionMethod.Examine, session, mailbox);
    }

    private bool SelectExamineCore(ImapTransactionMethod method, ImapSession session, ImapMailbox mailbox)
    {
      RejectNonAuthenticatedState(session);

      if (mailbox != null && session.SelectedMailbox == mailbox) {
        TraceVerbose(session, "mailbox has already been selected");
        return true;
      }

      var request = new ImapRequest(method, session.Timeout);

      if (mailbox != null) {
        ValidateMailboxRelationship(session, mailbox);

        if ((mailbox.MailboxFlags & ImapMailboxFlags.NoSelect) != 0)
          throw new InvalidOperationException("mailbox is not selectable");

        request.Parameters["Mailbox"] = mailbox.Name;
      }

      ImapResponse response;

      if (RequestSync(session, request, out response)) {
        session.Status = ImapSessionStatus.Selected;
        session.SelectedMailbox = (response as ImapMailboxResponse).Mailbox;

        TraceVerbose(session, "now in selected state (selected '{0}')", session.SelectedMailbox.Name);

        return true;
      }

      return false;
    }

    /// <summary>sends CREATE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Create(ImapSession session, string mailboxName)
    {
      RejectNonAuthenticatedState(session);

      throw new NotImplementedException();
    }

    /// <summary>sends DELETE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Delete(ImapSession session, ImapMailbox mailbox)
    {
      RejectNonAuthenticatedState(session);

      ValidateMailboxRelationship(session, mailbox);

      throw new NotImplementedException();
    }

    /// <summary>sends RENAME command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Rename(ImapSession session, ImapMailbox mailbox, string newMailboxName)
    {
      RejectNonAuthenticatedState(session);

      ValidateMailboxRelationship(session, mailbox);

      throw new NotImplementedException();
    }

    /// <summary>sends SUBSCRIBE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Subscribe(ImapSession session, ImapMailbox mailbox)
    {
      RejectNonAuthenticatedState(session);

      ValidateMailboxRelationship(session, mailbox);

      throw new NotImplementedException();
    }

    /// <summary>sends UNSUBSCRIBE command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Unsubscribe(ImapSession session, ImapMailbox mailbox)
    {
      RejectNonAuthenticatedState(session);

      ValidateMailboxRelationship(session, mailbox);

      throw new NotImplementedException();
    }

    /// <summary>sends LIST command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool List(ImapSession session, string mailboxName, out ImapMailbox[] mailboxes)
    {
      return ListLsubCore(ImapTransactionMethod.List, session, mailboxName, out mailboxes);
    }

    /// <summary>sends LSUB command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Lsub(ImapSession session, string mailboxName, out ImapMailbox[] mailboxes)
    {
      return ListLsubCore(ImapTransactionMethod.Lsub, session, mailboxName, out mailboxes);
    }

    private bool ListLsubCore(ImapTransactionMethod listMethod, ImapSession session, string mailboxName, out ImapMailbox[] mailboxes)
    {
      RejectNonAuthenticatedState(session);

      mailboxes = null;

      var request = new ImapRequest(listMethod, session.Timeout);

      request.Parameters["Mailbox"] = mailboxName;

      ImapResponse response;

      if (RequestSync(session, request, out response)) {
        mailboxes = (response as ImapListResponse).Mailboxes;
        return true;
      }

      return false;
    }

    /// <summary>sends STATUS command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Status(ImapSession session, ImapMailbox mailbox/*, TODO: Items*/)
    {
      RejectNonAuthenticatedState(session);

      ValidateMailboxRelationship(session, mailbox);

      var request = new ImapRequest(ImapTransactionMethod.Status, session.Timeout);

      request.Parameters["Mailbox"] = mailbox.Name;

      ImapResponse response;

      if (RequestSync(session, request, out response)) {
        // 常に同じインスタンスとなるはず
        return (mailbox == (response as ImapMailboxResponse).Mailbox);
      }

      return false;
    }

    /// <summary>sends APPEND command</summary>
    /// <remarks>valid in authenticated state</remarks>
    public bool Append(ImapSession session, ImapMailbox mailbox, ImapMessage message)
    {
      RejectNonAuthenticatedState(session);

      ValidateMailboxRelationship(session, mailbox);

      throw new NotImplementedException();
    }

    /*
     * requesting methods for selected state
     */

    /// <summary>sends CHECK command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Check(ImapSession session)
    {
      RejectNonSelectedState(session);

      return RequestSync(session, ImapTransactionMethod.Check);
    }

    /// <summary>sends CLOSE command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Close(ImapSession session)
    {
      RejectNonAuthenticatedState(session);

      if (session.Status != ImapSessionStatus.Selected) {
        TraceVerbose(session, "mailbox has already closed or has not selected yet");
        return true;
      }

      if (RequestSync(session, ImapTransactionMethod.Close)) {
        session.Status = ImapSessionStatus.Authenticated;
        session.SelectedMailbox = null;

        TraceVerbose(session, "now in authenticated state");

        return true;
      }

      return false;
    }

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Expunge(ImapSession session)
    {
      RejectNonSelectedState(session);

      throw new NotImplementedException();
    }

    /// <summary>sends SEARCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Search(ImapSession session, string query/*TODO:, ImapSearchQuery query*/, out int[] messages)
    {
      RejectNonSelectedState(session);

      messages = null;

      var request = new ImapRequest(ImapTransactionMethod.Search, session.Timeout);

      request.Parameters["Search-Query"] = query;

      ImapResponse response;

      if (RequestSync(session, request, out response)) {
        messages = (response as ImapMessageListResponse).Messages;
        return true;
      }

      return false;
    }

    /// <summary>sends UID SEARCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool UidSearch(ImapSession session, string query/*TODO:, ImapSearchQuery query*/, out int[] messages)
    {
      RejectNonSelectedState(session);

      throw new NotImplementedException();
    }

    /// <summary>sends FETCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Fetch(ImapSession session, int[] sequenceSet, string messageDataItems, out ImapMessage[] messages)
    {
      return FetchCore(session, GetFetchingSetString(sequenceSet), false, messageDataItems, out messages);
    }

    /// <summary>sends FETCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Fetch(ImapSession session, string sequenceSet, string messageDataItems, out ImapMessage[] messages)
    {
      return FetchCore(session, sequenceSet, false, messageDataItems, out messages);
    }

    /// <summary>sends UID FETCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool UidFetch(ImapSession session, int[] uidSet, string messageDataItems, out ImapMessage[] messages)
    {
      return FetchCore(session, GetFetchingSetString(uidSet), true, messageDataItems, out messages);
    }

    /// <summary>sends UID FETCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool UidFetch(ImapSession session, string uidSet, string messageDataItems, out ImapMessage[] messages)
    {
      return FetchCore(session, uidSet, true, messageDataItems, out messages);
    }

    private string GetFetchingSetString(int[] @set)
    {
      var setString = string.Empty;

      if (@set == null || @set.Length == 0) {
        // All
        setString = "1:*";
      }
      else {
        for (var i = 0; i < @set.Length; i++) {
          if (i != 0)
            setString += ",";

          setString += @set[i].ToString();
        }
      }

      return setString;
    }

    private bool FetchCore(ImapSession session, string fetchingSet, bool uidSet, string messageDataItems, out ImapMessage[] messages)
    {
      RejectNonSelectedState(session);

      messages = null;

      var request = new ImapRequest(ImapTransactionMethod.Fetch, session.Timeout);

      request.Parameters["Mailbox"] = session.SelectedMailbox.Name;

      // sequence set/message data item names
      if (string.Empty.Equals(messageDataItems))
        // All
        request.Parameters["Message-Data-Items"] = "ALL";
      else
        request.Parameters["Message-Data-Items"] = messageDataItems;

      var fetchingSetKey = (uidSet) ? "UID" : "Sequence-Set";

      if (string.Empty.Equals(fetchingSet))
        // All
        request.Parameters[fetchingSetKey] = "1:*";
      else 
        request.Parameters[fetchingSetKey] = fetchingSet;

      ImapResponse response;

      if (RequestSync(session, request, out response)) {
        messages = (response as ImapMessageResponse).Messages;

        foreach (var message in messages) {
          message.Mailbox = session.SelectedMailbox;
        }

        return true;
      }

      return false;
    }

    /// <summary>sends STORE command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Store(ImapSession session /*TODO:, ImapMessageFlags flags, ImapStoreOption(Add/Remove/Set) option */)
    {
      RejectNonSelectedState(session);

      throw new NotImplementedException();
    }

    /// <summary>sends UID STORE command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool UidStore(ImapSession session /*TODO:, ImapMessageFlags flags, ImapStoreOption(Add/Remove/Set) option */)
    {
      RejectNonSelectedState(session);

      throw new NotImplementedException();
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool Copy(ImapSession session, ImapMessage[] messages)
    {
      RejectNonSelectedState(session);

      throw new NotImplementedException();
    }

    /// <summary>sends UID COPY command</summary>
    /// <remarks>valid in selected state</remarks>
    public bool UidCopy(ImapSession session, ImapMessage[] messages)
    {
      RejectNonSelectedState(session);

      throw new NotImplementedException();
    }

    /*
     * requesting methods for authenticated/selected state
     */

    /// <summary>sends IDLE command</summary>
    /// <remarks>
    /// valid in authenticated/selected state,
    /// this method selects INBOX if no mailbox has been selected
    /// </remarks>
    public bool BeginIdle(ImapSession session, ImapIdlingResponseCallback callback)
    {
      RejectNonAuthenticatedState(session);

      return BeginIdle(session, null, callback);
    }

    /// <summary>sends IDLE command</summary>
    /// <remarks>
    /// valid in authenticated/selected state,
    /// this method selects mailbox if no mailbox has been selected
    /// </remarks>
    public bool BeginIdle(ImapSession session, ImapMailbox mailbox, ImapIdlingResponseCallback callback)
    {
      RejectNonAuthenticatedState(session);

      if (session.Connection.IsIdling)
        throw new InvalidOperationException("already idling");

      if (session.SelectedMailbox != mailbox) {
        if (session.SelectedMailbox != null) {
          // close if selected
          if (!Close(session))
            return false;
        }

        if (mailbox != null) {
          if (!Select(session, mailbox))
            return false;
        }
      }

      session.Connection.IdlingCallback = callback;

      var request = new ImapRequest(ImapTransactionMethod.Idle, session.Timeout);

      if (session.SelectedMailbox != null)
        request.Parameters["Mailbox"] = session.SelectedMailbox.Name;

      var succeeded = true;
      var transaction = RequestAsync(session, request);

      for ( ; ; Thread.Sleep(50)) {
        if (session.Connection.IsIdling)
          break; // succeeded
        if (transaction.Response != null &&
            transaction.Response.Status != ImapTransactionStatus.Ok) {
          succeeded = false; // failed
          break;
        }
        if (request.Timeout < DateTime.Now.Subtract(request.RequestedTime).TotalMilliseconds) {
          succeeded = false; // time out
          break;
        }
      }

      if (succeeded)
        TraceVerbose(session, "session is idling now");
      else
        session.Connection.IdlingCallback = null;

      return succeeded;
    }

    /// <summary>sends DONE for IDLE command continuation request responses</summary>
    /// <remarks>valid in idling</remarks>
    public bool EndIdle(ImapSession session)
    {
      RejectNonAuthenticatedState(session);

      if (!session.Connection.IsIdling) {
        TraceVerbose(session, "not idled");
        return true;
      }

      var request = new ImapRequest(ImapTransactionMethod.IdleDone, session.Timeout);

      if (!RequestSync(session, request))
        return false; // failed

      var succeeded = true;

      for ( ; ; Thread.Sleep(50)) {
        if (!session.Connection.IsIdling)
          break; // succeeded
        if (request.Timeout < DateTime.Now.Subtract(request.RequestedTime).TotalMilliseconds) {
          succeeded = false;  // time out
          break;
        }
        agent.Signal();
      }

      if (succeeded) {
        TraceVerbose(session, "idle completed");
        session.Connection.IdlingCallback = null;
      }

      return succeeded;
    }

    /// <summary>sends NAMESPACE command</summary>
    /// <remarks>valid in authenticated/selected state</remarks>
    public bool Namespace(ImapSession session, out ImapNamespace @namespace)
    {
      RejectNonAuthenticatedState(session);

      @namespace = null;
      
      ImapResponse response;

      if (RequestSync(session, ImapTransactionMethod.Namespace, out response)) {
        @namespace = (response as ImapNamespaceResponse).Namespace;

        return true;
      }

      return false;
    }

    /*
     * other requesting methods
     */

    /// <summary>sends XYZZY command</summary>
    /// <remarks>Gimap XYZZY capability extension</remarks>
    public bool Xyzzy(ImapSession session)
    {
      RejectNonConnectedState(session);

      return RequestSync(session, ImapTransactionMethod.Xyzzy);
    }

    /*
     * internal requesting methods
     */
    private ImapTransaction RequestAsync(ImapSession session, ImapRequest request)
    {
      RejectDisposed();

      if (request.Method != ImapTransactionMethod.IdleDone)
        RejectIdling(session);

      // TODO: レスポンスがDisconnectingになった場合
      return agent.Request(session.Connection, request);
    }

    private bool RequestSync(ImapSession session, ImapTransactionMethod method)
    {
      ImapResponse response;

      return RequestSync(session, method, out response);
    }

    private bool RequestSync(ImapSession session, ImapTransactionMethod method, out ImapResponse response)
    {
      return RequestSync(session, new ImapRequest(method, session.Timeout), out response);
    }

    private bool RequestSync(ImapSession session, ImapRequest request)
    {
      ImapResponse response;

      return RequestSync(session, request, out response);
    }

    private bool RequestSync(ImapSession session, ImapRequest request, out ImapResponse response)
    {
      RejectDisposed();

      response = null;

      if (request.Method != ImapTransactionMethod.IdleDone)
        RejectIdling(session);

      for (var retry = 0; retry < 3; retry++) {
        var transaction = agent.Request(session.Connection, request);

        while (transaction.Response == null) {
          Thread.Sleep(20);
        }

        response = transaction.Response;

        if (response.Status == ImapTransactionStatus.TransactionProgress) {
          // retry
          Thread.Sleep(250);
          continue;
        }
        else {
          break;
        }
      }

      if (response.Status == ImapTransactionStatus.Disconnecting) {
        session.Status = ImapSessionStatus.NonAuthenticated;

        TraceVerbose(session, "now in non-authenticated state");

        Disconnect(session);
      }

      // true for 2xx, false for 3xx/4xx/5xx
      return ((int)response.Status < 300);
    }

    /*
     * internal methods
     */

    internal ImapSession GetExistSession(ImapUri uri)
    {
      lock (sessions) {
        if (sessions.ContainsKey(uri.UserAtHost))
          return sessions[uri.UserAtHost];
      }

      return null;
    }

    /*
     * methods for checking status
     */

    private static void RejectNonConnectedState(ImapSession session)
    {
      if (session == null)
        throw new NullReferenceException("session is null");

      if (session.Status == ImapSessionStatus.NotConnected)
        throw new InvalidOperationException("session has been disconnected");
    }

    private static void RejectNonAuthenticatedState(ImapSession session)
    {
      RejectNonConnectedState(session);

      if (session.Status == ImapSessionStatus.NonAuthenticated)
        throw new InvalidOperationException("session is not in authenticated state");
    }

    private static void RejectNonSelectedState(ImapSession session)
    {
      RejectNonAuthenticatedState(session);

      if (session.Status != ImapSessionStatus.Selected)
        throw new InvalidOperationException("session is not in selected state");
    }

    private static void RejectIdling(ImapSession session)
    {
      if (session.Connection.IsIdling)
        throw new InvalidOperationException("connection is idling now");
    }

    private static void ValidateMailboxRelationship(ImapSession session, ImapMailbox mailbox)
    {
      if (mailbox.ServerInfo != session.Connection.ServerInfo)
        throw new InvalidOperationException("invalid mailbox");
    }

    /*
     * private methods
     */

    private void Disconnect(ImapSession session)
    {
      lock (sessions) {
        if (sessions.ContainsKey(session.HostUri.UserAtHost)) {
          sessions.Remove(session.HostUri.UserAtHost);
        }
      }

      session.Close();

      TraceVerbose(session, "now in non-connected state");
    }

    private Stream CreateAuthenticatedStream(ImapUri uri, Stream baseStream)
    {
      if (createAuthenticatedStreamCallback != null) {
        return createAuthenticatedStreamCallback(uri, baseStream);
      }
      else {
        const int imapsPort = 993;

        if (uri.Port == imapsPort) {
          SslStream sslStream = null;

          if (serverCertificateValidationCallback == null)
            sslStream = new SslStream(baseStream, false);
          else
            sslStream = new SslStream(baseStream, false, serverCertificateValidationCallback);

          sslStream.AuthenticateAsClient(uri.Host, certCollection, SslProtocols.Default, true);

          Trace.Verbose("[ImapSession] Created SSL stream for {0}", uri);
          Trace.Verbose("[ImapSession]   IsAuthenticated: {0}", sslStream.IsAuthenticated);
          Trace.Verbose("[ImapSession]   IsMutuallyAuthenticated: {0}", sslStream.IsMutuallyAuthenticated);
          Trace.Verbose("[ImapSession]   IsEncrypted: {0}", sslStream.IsEncrypted);
          Trace.Verbose("[ImapSession]   IsSigned: {0}", sslStream.IsSigned);

          return sslStream;
        }
        else {
          return baseStream;
        }
      }
    }

    [Conditional("TRACE")]
    private void TraceVerbose(ImapSession session, string format, params object[] arguments)
    {
      Trace.Verbose("CID:{0} （≧▽≦）[ImapSession] {1}", session.Id, string.Format(format, arguments));
    }

    private void RejectDisposed()
    {
      if (disposed)
        throw new ObjectDisposedException(GetType().Name);
    }

    private bool disposed = false;
    private ImapTransactionAgent agent = null;
    private readonly Dictionary<string, ImapSession> sessions = new Dictionary<string, ImapSession>();
    private readonly X509Certificate2Collection certCollection = new X509Certificate2Collection();
    private RemoteCertificateValidationCallback serverCertificateValidationCallback = null;
    private ImapCreateAuthenticatedStreamCallback createAuthenticatedStreamCallback = null;
  }
}