using System;
using System.Collections.Generic;
using System.Threading;

using Smdn.Protocols.Imap4.Protocol;

namespace Smdn.Protocols.Imap4.Transaction {
  // Layers:
  //   Client.ImapClient
  //   Client.ImapClientAgent
  //   Session.ImapSessionAgent
  //   Session.ImapSession
  // * Transaction.ImapTransactionAgent
  //   Transaction.ImapTransaction/Transaction.ImapRequest/Transaction.ImapResponse
  //   Protocol.ImapConnection/Protocol.ImapProtocolCommand/Protocol.ImapProtocolResponse
  //   (IMAP4)

  public class ImapTransactionAgent : IDisposable {
    /*
     * construction/destruction
     */

    public ImapTransactionAgent()
    {
    }

    public void Initialize()
    {
      if (agentThread != null)
        throw new InvalidOperationException("already initialized");

      Trace.Info("ImapTransactionAgent initializing");

      agentThread = new Thread((ThreadStart)AgentThread);
      agentThread.IsBackground = true;
      agentThread.Start();

      // wait for initial WaitAny blocking
      for (;; Thread.Sleep(100)) {
        if ((agentThread.ThreadState & ThreadState.WaitSleepJoin) != 0)
          break;
      }

      Trace.Info("ImapTransactionAgent initialized");
    }

    ~ImapTransactionAgent()
    {
      Dispose();
      GC.SuppressFinalize(this);
    }

    public void Dispose()
    {
      if (disposed)
        return;

      Trace.Info("ImapTransactionAgent disposing");

      abortedEvent.Set();

      if (agentThread != null) {
        agentThread.Join();
        agentThread = null;
      }

      Trace.Info("ImapTransactionAgent disposed");

      disposed = true;
    }

    /*
     * methods for requesting
     */

    public ImapTransaction Request(ImapConnection connection, ImapRequest request)
    {
      if (abortedEvent.WaitOne(0, false))
        throw new InvalidOperationException("aborting");

      var transaction = ImapTransaction.Create(connection, request);

      transaction.Request.RequestedTime = DateTime.Now;

      // check whether another transaction on this connection is in progress
      if (request.Method != ImapTransactionMethod.IdleDone) {
        foreach (var t in processingTransactions.ToArray()) {
          // TODO: '5.5. Multiple Commands in Progress'
          //   The following are examples of valid non-waiting command sequences:
          //      FETCH + STORE + SEARCH + CHECK
          //      STORE + COPY + EXPUNGE
          if (t.Connection == connection) {
            transaction.Response = new ImapResponse(ImapTransactionStatus.TransactionProgress,
                                                    "another transaction on this connection is in progress");
            return transaction;
          }
        }
      }

      // enqueue
      lock (processingTransactions) {
        processingTransactions.Add(transaction);
      }

      Trace.Log(request, transaction);

      Signal();

      return transaction;
    }

    /// <summary>sends a activation signal to transaction processing thread</summary>
    public void Signal()
    {
      requestedEvent.Set();
    }

    /*
     * methods for transaction processing thread
     */

    private void AgentThread()
    {
      Trace.Info("ImapTransactionAgent thread started");

      try {
        var waitEvents = new WaitHandle[] {abortedEvent, requestedEvent};

        abortedEvent.Reset();

        // wait for first request (or abort request)
        WaitHandle.WaitAny(waitEvents); 

        for (;;) {
          ImapTransaction[] transactions;

          lock (processingTransactions) {
            transactions = processingTransactions.ToArray();
          }

          int idlingTransactions;
          bool aborted;

          ProcessTransactions(transactions, out aborted, out idlingTransactions);

          if (aborted)
            break;

          if (0 < processingTransactions.Count) {
            if (idlingTransactions < processingTransactions.Count)
              // アイドル中ではないトランザクションがある場合は、すぐに処理する
              Thread.Sleep(10);
            else
              // アイドル中のトランザクションのみの場合は、次のリクエストが来る
              // までしばらく待機する
              requestedEvent.WaitOne(10000, false);
          }
          else {
            // 処理すべきトランザクションが無い場合は、リクエストが来るか、
            // スレッドの停止が要求されるまで待機する
            var index = WaitHandle.WaitAny(waitEvents); 

            if (index == 0)
              // aborted
              break;
          }
        }

        // abort all transactions
        foreach (var transaction in processingTransactions) {
          transaction.Abort();
        }

        processingTransactions.Clear();
      }
      catch (ThreadAbortException) {
        Trace.Info("ImapTransactionAgent thread force aborted");
      }
      catch (Exception ex) {
        Trace.Log(ex);
        throw ex;
      }

      Trace.Info("ImapTransactionAgent thread finished");
    }

    private void ProcessTransactions(ImapTransaction[] transactions, out bool aborted, out int idlingTransactions)
    {
      idlingTransactions = 0;
      aborted = false;

      foreach (var transaction in transactions) {
        // aborted
        if (abortedEvent.WaitOne(0, false)) {
          aborted = true;
          return;
        }

        if (ProcessTransaction(transaction)) {
          // transaction is finished
          lock (processingTransactions) {
            processingTransactions.Remove(transaction);
          }
        }
        else if (transaction.Connection.IsIdling) {
          idlingTransactions++;
        }
      }
    }

    private bool ProcessTransaction(ImapTransaction transaction)
    {
      transaction.Process();

      if (transaction.Response != null) {
        // transaction is finished
        Trace.Log(transaction.Response, transaction);
        return true;
      }

      // transaction is not finished
      if (transaction.Connection.IsIdling)
        return false;

      if (transaction.Request.Timeout < 0)
        return false;

      if (transaction.Request.Timeout <
          DateTime.Now.Subtract(transaction.Request.RequestedTime).TotalMilliseconds)
        // request time out
        transaction.Abort();

      return false;
    }

    private bool disposed = false;
    private readonly List<ImapTransaction> processingTransactions = new List<ImapTransaction>();
    private Thread agentThread = null;
    private readonly AutoResetEvent requestedEvent = new AutoResetEvent(false);
    private readonly ManualResetEvent abortedEvent = new ManualResetEvent(false);
  }
}
