using System;
using System.Collections.Generic;

using Smdn.Protocols.Imap4.Protocol;

namespace Smdn.Protocols.Imap4.Transaction.Transactions {
  internal abstract class SelectTransactionBase : TransactionBase {

    // ProcessParameterNotSetted
    //   x Conclude
    // ProcessMailboxNotFound
    //   x Conclude
    // ProcessSelect
    //   ->WaitForSelect
    //   x Conclude
    // WaitForSelect
    //   x Conclude

    protected SelectTransactionBase(string command, ImapConnection connection, ImapRequest request)
      : base(connection, request)
    {
      commandString = command.ToLower();
    }

    protected override ProcessTransactionDelegate Reset()
    {
      if (!Request.Parameters.ContainsKey("Mailbox"))
        return ProcessParameterNotSetted;

      selectingMailbox = ImapMailboxManager.GetExistMailbox(Connection.ServerInfo, Request.Parameters["Mailbox"]);

      if (selectingMailbox == null)
        // TODO: LIST/LSUBで既知のメールボックスに対してしかSELECTできない
        return ProcessMailboxNotFound;

      return ProcessSelect;
    }

    private void ProcessParameterNotSetted()
    {
      Conclude(new ImapResponse(ImapTransactionStatus.BadRequest, "parameter 'Mailbox' must be setted"));
    }

    private void ProcessMailboxNotFound()
    {
      Conclude(new ImapResponse(ImapTransactionStatus.NotFound, string.Format("mailbox '{0}' was not found", Request.Parameters["Mailbox"])));
    }

    private void ProcessSelect()
    {
      lastTag = Send(commandString, string.Format("\"{0}\"", selectingMailbox.Name));

      Transit(WaitForSelect);
    }

    private void WaitForSelect()
    {
      ImapProtocolResponse response;

      if (!Receive(out response))
        return;

      if (response is ImapDataResponse) {
        ImapMessageFlags? flags;
        string[] extendedFlags;
        int? exists;
        int? recent;
        if (ImapDataResponseConverter.FromSelectExpungeCommandResponse(response as ImapDataResponse,
                                                                       out flags,
                                                                       out extendedFlags,
                                                                       out exists,
                                                                       out recent)) {
          if (flags != null)          selectingMailbox.ApplicableFlags = (ImapMessageFlags)flags;
          if (extendedFlags != null)  selectingMailbox.ExtendedApplicableFlags = extendedFlags;
          if (exists != null)         selectingMailbox.ExistsMessage = (int)exists;
          if (recent != null)         selectingMailbox.RecentMessage = (int)recent;
        }
      }
      else if (response is ImapStatusResponse) {
        var status = response as ImapStatusResponse;

        if (status.ResponseCode != null) {
          bool? readOnly;
          int? unseen;
          ImapMessageFlags? permanentFlags;
          string[] extendedPermanentFlags;
          int? uidNext;
          int? uidValidity;

          if (ImapStatusResponseConverter.FromSelectExamineResponse(status,
                                                                    out readOnly,
                                                                    out unseen,
                                                                    out permanentFlags,
                                                                    out extendedPermanentFlags,
                                                                    out uidNext,
                                                                    out uidValidity)) {
            if (readOnly != null)               selectingMailbox.ReadOnly = (bool)readOnly;
            if (unseen != null)                 selectingMailbox.UnseenMessage = (int)unseen;
            if (permanentFlags != null)         selectingMailbox.PermanentFlags = (ImapMessageFlags)permanentFlags;
            if (extendedPermanentFlags != null) selectingMailbox.ExtendedPermanentFlags = extendedPermanentFlags;
            if (uidNext != null)                selectingMailbox.UidNext = (int)uidNext;
            if (uidValidity != null)            selectingMailbox.UidValidity = (int)uidValidity;
          }
        }

        if (status.Tag != lastTag)
          return;

        if (status.Status == ImapProtocolStatus.OK) {
          var description = "selected mailbox";

          if (selectingMailbox.ReadOnly) {
            description += " as read-only";
          }

          Trace.Dump(selectingMailbox);

          Conclude(new ImapMailboxResponse(ImapTransactionStatus.Ok, selectingMailbox, description));
        }
        else {
          ConcludeError(status);
        }
      }
    }

    private string lastTag = string.Empty;
    private string commandString = string.Empty;
    private ImapMailbox selectingMailbox = null;
  }
}
